/*********************************************************************
 *
 * file name : ipconfig.c
 * summary   : program to configure network information of AJ
 * coded by  : mori
 * copyright : Atmark Techno
 * date(init): 2003.12.04
 * 
 ********************************************************************/

#include <iostream>
#include <string.h>
#include <cstdlib>
#include <cstdio>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>
#include <net/route.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <errno.h>

#include "discover_lib.h"

/* MACRO define */
#define REPLY_TIMEOUT	10	// 5s
#define TO_INADDR(x)	(*(struct in_addr*)&x)
#define OUTPUT_FORMAT	"%02x:%02x:%02x:%02x:%02x:%02x|DHCP %-3s|IP %-15s|MASK %-15s|GW %-15s"

using namespace std;

enum error_reason_e
{
	ENOERROR = 0,
	ECOMMAND,
	ESOCKET,
	ESENDPACKET,
	ERECVPACKET,
	EREPLY,
};

/**********************************************************************
 * summary : Output network info to find
 * return  :
 *********************************************************************/
void usage()
{
	printf("usage:\n");
	printf("ipconfig MAC                  - DHCP\n");
	printf("ipconfig MAC [IP][subnet]     - Static ip without GW\n");
	printf("ipconfig MAC [IP][subnet][GW] - Static ip with GW\n");
	printf("ex:\n");
	printf("ipconfig 12:34:56:78:9a:bc 192.168.0.100 255.255.255.0 192.168.0.1\n");
	return;
}

/**********************************************************************
 * summary : Output network info to find
 * return  :
 *********************************************************************/
void PrintReply(const discover_packet_s* packet)
{
	char buff[512];
	char ip_address[16];
	char netmask[16];
	char gateway[16];

	strcpy(ip_address, inet_ntoa(TO_INADDR(packet->net_info.ip_address)));
	strcpy(netmask, inet_ntoa(TO_INADDR(packet->net_info.netmask)));
	strcpy(gateway, inet_ntoa(TO_INADDR(packet->net_info.gateway)));

	sprintf(
			buff,
			OUTPUT_FORMAT,
			(unsigned char)packet->mac_addr[0],
			(unsigned char)packet->mac_addr[1],
			(unsigned char)packet->mac_addr[2],
			(unsigned char)packet->mac_addr[3],
			(unsigned char)packet->mac_addr[4],
			(unsigned char)packet->mac_addr[5],
			packet->net_info.use_dhcp? "yes" : "no",
			ip_address,
			netmask,
			gateway
			);
	cout << buff << endl;
}

/**********************************************************************
 * summary : main routine
 * return  : success -> 0, failed -> !0
 *********************************************************************/
bool SetNetworkInfo(
	network_info_s* netinfo,
	const char* ipaddr,
	const char* subnet,
	const char* gateway
)
{
	if(ipaddr == NULL){
		/* DHCP */
		netinfo->use_dhcp = htonl(1);
		return true;
	}
	
	if(subnet == NULL){
		return false;
	}
	if(!inet_aton(ipaddr, (struct in_addr*)&netinfo->ip_address)){
		return false;
	}
	if(!inet_aton(subnet, (struct in_addr*)&netinfo->netmask)){
		return false;
	}

	unsigned int mask;
	int workAddr = htonl(netinfo->ip_address);
	if(!(workAddr & (1 << 31))){
		mask = 0xff000000;
	}
	else if(!(workAddr & (1 << 30))){
		mask = 0xffff0000;
	}
	else if(!(workAddr & (1 << 29))){
		mask = 0xffffff00;
	}
	else{
		return false;
	}

	if((workAddr | mask) == mask ||
		(workAddr | mask) == 0xffffffff){
		return false;
	}

	if((netinfo->ip_address | netinfo->netmask) == netinfo->netmask ||
		(netinfo->ip_address | netinfo->netmask) == 0xffffffff){
		return false;
	}

	if(gateway == NULL){
		/* no GW */
		netinfo->gateway = 0;
		return true;
	}

	return inet_aton(gateway, (struct in_addr*)&netinfo->gateway);
}

/**********************************************************************
 * summary : main routine
 * return  : success -> 0, failed -> !0
 *********************************************************************/
bool SetMacAddress(char* macaddr, const char* macaddr_str)
{
	if(strlen(macaddr_str) != 17){
		return false;
	}

	char hexnum[3];
	memset(hexnum, 0, sizeof(hexnum));

	for(int i = 0 ; i < 6 ; i++){
		char* err;
		strncpy(hexnum, macaddr_str + (i * 3), 2);
		macaddr[i] = (unsigned int)strtol(hexnum, &err, 16);
		if(*err){
			return false;
		}
	}

	return true;
}

/**********************************************************************
 * summary : main routine
 * return  : success -> 0, failed -> !0
 *********************************************************************/
int main(int argc, char **argv)
{
	char macAddr[6];

	/* get arguments */
	if(argc == 1){
		usage();
		return -ECOMMAND;
	}

	if(!SetMacAddress(macAddr, argv[1])){
		usage();
		return -ECOMMAND;
	}
	
	network_info_s networkInfo;
	memset(&networkInfo, 0, sizeof(networkInfo));

	switch(argc){
	case 2:
		/* using dhcp */
		if(!SetNetworkInfo(&networkInfo, argv[2], NULL, NULL)){
			usage();
			return -ECOMMAND;
		}
		break;
	case 4:
		/* static IP without GW */
		if(!SetNetworkInfo(&networkInfo, argv[2], argv[3], NULL)){
			printf("Invalid IP setting\n");
			usage();
			return -ECOMMAND;
		}
		break;
	case 5:
		/* static IP with GW */
		if(!SetNetworkInfo(&networkInfo, argv[2], argv[3], argv[4])){
			printf("Invalid IP setting\n");
			usage();
			return -ECOMMAND;
		}
		break;
	default:
		usage();
		return -ECOMMAND;
	}
	
	CDiscover discover;
	if(!discover.Setup()){
		cerr << "failed to prepare broadcast socket" << endl;
		return -ESOCKET;
	}

	LOG("sending network config request\n");
	uint32_t transactionID = discover.SendConfigPacket(
											macAddr, &networkInfo);
	if(!transactionID){
		cerr << "failed to send request packet" << endl;
		return -ESENDPACKET;
	}

	discover_packet_s packet;
	memset(&packet, 0, sizeof(packet));
	LOG("wait for reply\n");
	if(!discover.RecvReplyPacket(&packet, transactionID, REPLY_TIMEOUT)){
		cerr << "failed to receive reply packet" << endl;
		return -ERECVPACKET;
	}

	cout << "packet received" << endl;
	if(packet.message_type != ntohl(IP_OFFER_REP)){
		cerr << "other message type's reply received??\n" << endl;
		return -ERECVPACKET;
	}
	if(packet.result != ntohl(SUCCESS)){
		cerr << "ERROR! reason : " << ntohl(packet.result) << endl;
		return -EREPLY;
	}

	PrintReply(&packet);
	return ENOERROR;
}

