//////////////////////////////////////////////////////////////////////
//
// file name : CJBLib.cpp
// summary   : JTAG-Blazer network transfer library
//
//////////////////////////////////////////////////////////////////////

#ifdef WIN32
	#include <afxsock.h>
	#define	LOG		TRACE
	#define	ERR		TRACE
#else
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <fcntl.h>
	#include <unistd.h>
	#include <errno.h>
#ifdef __DEBUG__
	#define	LOG(args...)		printf(args)
	#define	ERR(args...)		printf(args)
	#define ASSERT(args...)		while(0){}
#else
	#define	LOG(args...)	
	#define	ERR(args...)
	#define ASSERT(args...)	
#endif
#endif

#include <stdio.h>
#include <sys/stat.h>
#include "jblib.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define PROTOCOL_VERSION		1
#define NORMAL_PORT_NO			39293
#define UPDATE_PORT_NO			39793
#define WRITE_BUF_SIZE			1400
#define CHECK_SUM_SIZE			4
#define CONNECT_TIMEOUT_SEC		1
#define CONNECT_TIMEOUT_USEC	0

//--------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////
// summary : constructor with IP
// return  : none
//////////////////////////////////////////////////////////////////////
CJBLib::CJBLib(unsigned long targetIP)
{
	m_socket	= -1;
	m_targetIP	= targetIP;
	m_portNo	= NORMAL_PORT_NO;
}

//////////////////////////////////////////////////////////////////////
// summary : destructor
// return  : none
//////////////////////////////////////////////////////////////////////
CJBLib::~CJBLib()
{
	CloseSocket();
}

//////////////////////////////////////////////////////////////////////
// summary :
// return  :
//////////////////////////////////////////////////////////////////////
bool CJBLib::PrepareSocket(
	unsigned long targetIP,
	unsigned short portNo
)
{
	LOG("PrepareSocket ip : %08x, port : %d\n", targetIP, portNo);

	bool bReconnect = false;
	if(targetIP && m_targetIP != targetIP){
		bReconnect = true;
		m_targetIP = targetIP;
	}
	if(portNo && m_portNo != portNo){
		bReconnect = true;
		m_portNo = portNo;
	}

	if(m_socket >= 0 && !bReconnect){
		// already connected
		ERR("socket already connected\n");
		return true;
	}

	CloseSocket();

	m_socket = ::socket(PF_INET, SOCK_STREAM, 0);
	if(m_socket < 0){
		ERR("socket() failed\n");			
		return false;
	}

	struct sockaddr_in sendAddr;
	memset((char*)&sendAddr, 0, sizeof(sendAddr));
	sendAddr.sin_family			= PF_INET;
	sendAddr.sin_port			= htons(m_portNo);

	bool result = true;
#ifdef WIN32
	sendAddr.sin_addr.s_addr	= htonl(m_targetIP);
	unsigned long opt = 1;
	ioctlsocket(m_socket, FIONBIO, &opt);
	if(::connect(m_socket, (struct sockaddr*)&sendAddr, sizeof(sendAddr)) < 0){
		if(WSAGetLastError() != WSAEWOULDBLOCK){
			ERR("connect() failed\n");			
			result = false;
			goto done;
		}
	}
#else
	sendAddr.sin_addr.s_addr	= m_targetIP;
	fcntl(m_socket, F_SETFL, O_NONBLOCK);
	if(::connect(m_socket, (struct sockaddr*)&sendAddr, sizeof(sendAddr)) < 0){
		if(errno != EINPROGRESS){
			LOG("connect() %d\n", errno);
			result = false;
			goto done;
		}
	}
#endif

	struct timeval timeout;
	timeout.tv_sec = CONNECT_TIMEOUT_SEC;
	timeout.tv_usec = CONNECT_TIMEOUT_USEC;

	fd_set writefds;
	FD_ZERO(&writefds);
	FD_SET(m_socket, &writefds);

	int ret;
	ret = ::select(m_socket + 1, NULL, &writefds, NULL, &timeout);
	if(ret <= 0){
		ERR("select() failed\n");			
		result = false;
		goto done;
	}

	unsigned long err;
	socklen_t errSize;
	errSize = sizeof(err);
	if(getsockopt(m_socket, SOL_SOCKET, SO_ERROR, (char*)&err, &errSize)){
		ERR("getsockopt() failed\n");			
		result = false;
		goto done;
	}
	if(err){
		ERR("getsockopt() SO_ERROR : %d\n", err);			
		result = false;
		goto done;
	}

done:
#ifdef WIN32
	opt = 0;
	ioctlsocket(m_socket, FIONBIO, &opt);
#else
	fcntl(m_socket, F_SETFL, 0);
#endif

	return result;
}

//////////////////////////////////////////////////////////////////////
// summary :
// return  :
//////////////////////////////////////////////////////////////////////
bool CJBLib::CancelRequest()
{
	if(m_socket < 0){
		return false;
	}

	CloseSocket();
	return true;
}

//////////////////////////////////////////////////////////////////////
// summary :
// return  :
//////////////////////////////////////////////////////////////////////
void CJBLib::CloseSocket()
{
	if(m_socket < 0){
		return;
	}

#ifdef WIN32
	::shutdown(m_socket, 2);
	::closesocket(m_socket);
#else
	::shutdown(m_socket, SHUT_RDWR);
	::close(m_socket);
#endif
	m_socket = -1;
}

//////////////////////////////////////////////////////////////////////
// summary : send 32bit data by network byte order
// return  : true / false
//////////////////////////////////////////////////////////////////////
bool CJBLib::SendLongData(
	unsigned long sendData
)
{
	unsigned long netData = htonl(sendData);
	int size = ::send(m_socket, (char*)&netData, sizeof(netData), 0);
	return size == sizeof(netData);
}

//////////////////////////////////////////////////////////////////////
// summary : receive 32bit data and return host byte order
// return  : true / false
//////////////////////////////////////////////////////////////////////
bool CJBLib::ReceiveLongData(
	unsigned long* recvData
)
{
	unsigned long netData;
	int size = ::recv(m_socket, (char*)&netData, sizeof(netData), 0);
	if(size !=	sizeof(netData)){
		return false;
	}
	*recvData = ntohl(netData);
	return true;
}

//////////////////////////////////////////////////////////////////////
// summary : receive responce header
// return  : error code
//////////////////////////////////////////////////////////////////////
bool CJBLib::ReceiveHeader(
	JBDataHeader* header
)
{
	unsigned long data[3];
	int size = ::recv(m_socket, (char*)data, sizeof(data), 0);
	if(size != sizeof(data)){
		return false;
	}

	header->dataID = (JBDataID)ntohl(data[0]);
	header->version = ntohl(data[1]);
	header->sequence = ntohl(data[2]);

	return true;
}

//////////////////////////////////////////////////////////////////////
// summary : receive responce header and status
// return  : error code
//////////////////////////////////////////////////////////////////////
JBErrorCode CJBLib::ReceiveResponse(
	JBDataID id
)
{
	JBDataHeader header;
	if(!ReceiveHeader(&header)){
		return JTAG_RECV_SOCKET_FAILED;
	}

	if(header.dataID != id ||
		header.sequence != m_sequence){
		return JTAG_INVALID_RESPONSE;
	}

	/* no need to check because server does */
	/*
	if(header.version != PROTOCOL_VERSION){
		return JTAG_VERSION_MISMATCH;
	}
	*/

	// status
	JBErrorCode errCode;
	ReceiveLongData((unsigned long*)&errCode);
	return errCode;
}

//////////////////////////////////////////////////////////////////////
// summary : make new sequence number
// return  : unique sequence number
//////////////////////////////////////////////////////////////////////
unsigned long CJBLib::GetNewSequenceNo()
{
	static unsigned long sequenceNo;
	static bool init = false;
	if(!init){
		srand((unsigned)time(NULL));
		sequenceNo = rand();
		init = true;
	}
	return sequenceNo++;
}

//////////////////////////////////////////////////////////////////////
// summary : receive responce common part
// return  : true / false
//////////////////////////////////////////////////////////////////////
bool CJBLib::SendHeader(
	JBDataID id
)
{
	m_sequence = GetNewSequenceNo();

	JBDataHeader header;
	header.dataID	= (JBDataID)htonl((unsigned long)id);
	header.version	= htonl((unsigned long)PROTOCOL_VERSION);
	header.sequence	= htonl(m_sequence);

	int size = ::send(m_socket, (char*)&header, sizeof(header), 0);
	return size == sizeof(header);
}

//--------------------------------------------------------------------
//====================================================================
// summary : write config data to target directory
// return  : error code
//====================================================================
JBErrorCode CJBLib::TargetWriteRequest(
	const char* fileName,
	bool littleBear,
	void (*ProgressNotify)(int status, int param)
)
{
	if(!PrepareSocket(0, NORMAL_PORT_NO)){
		return JTAG_OPEN_SOCKET_FAILED;
	}

	// send id
	if(!SendHeader(NET_WRITE_REQ)){
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// send mode
	if(!SendLongData((unsigned long)littleBear)){
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// send file it self
	return SendFileData(
					fileName,
					NET_WRITE_RES,
					NET_WRITE_NOTIFY,
					ProgressNotify
					);
}

//====================================================================
// summary : send download file request to write to flash rom
// return  : error code
//====================================================================
JBErrorCode CJBLib::FlashWriteRequest(
	const char* fileName,
	const FlashFileData* fileData,
	void (*ProgressNotify)(int status, int param)
)
{
	if(!PrepareSocket(0, NORMAL_PORT_NO)){
		return JTAG_OPEN_SOCKET_FAILED;
	}

	// send header
	if(!SendHeader(FLASH_WRITE_REQ)){
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// send file data
	JBErrorCode status = SendFlashFileData(fileData);
	if(status != JTAG_NO_ERROR){
		CloseSocket();
		return status;
	}

	// send file it self
	return SendFileData(
					fileName,
					FLASH_WRITE_RES,
					FLASH_WRITE_NOTIFY,
					ProgressNotify
					);
}

//--------------------------------------------------------------------
//====================================================================
// summary : write config data to target directory
// return  : error code
//====================================================================
JBErrorCode CJBLib::UpdateProgramRequest(
	const char* fileName,
	void (*ProgressNotify)(int status, int param)
)
{
	if(!PrepareSocket(0, UPDATE_PORT_NO)){
		return JTAG_OPEN_SOCKET_FAILED;
	}

	// send header
	if(!SendHeader(UPDATE_PROG_REQ)){
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// send file it self
	return SendFileData(
					fileName,
					UPDATE_PROG_RES,
					UPDATE_PROG_NOTIFY,
					ProgressNotify
					);
}

//////////////////////////////////////////////////////////////////////
// summary :
// return  :
//////////////////////////////////////////////////////////////////////
JBErrorCode CJBLib::SendFlashFileData(
	const FlashFileData* fileData
)
{
	LOG("SendFlashFileData()\n");

	size_t nameLen = strlen(fileData->fileName);
	unsigned long data[5];
        ssize_t sent;
	data[0] = htonl((unsigned long)fileData->littleBear);
	data[1] = htonl((unsigned long)fileData->compressed);
	data[2] = htonl((unsigned long)fileData->recordTime);
	data[3] = htonl(fileData->fileSize);
	data[4] = htonl(nameLen);

	int size = ::send(m_socket, (char*)data, sizeof(data), 0);
	if(size != sizeof(data)){
		return JTAG_SEND_SOCKET_FAILED;
	}
		
	// send file name
	LOG("file name : %s\n", fileData->fileName);
        sent = ::send(m_socket, fileData->fileName, nameLen, 0);
        if ((sent < 0) || (nameLen != (size_t)sent)) {
		return JTAG_SEND_SOCKET_FAILED;
	}

	LOG("Every sending data has done\n");
	return JTAG_NO_ERROR;
}

//////////////////////////////////////////////////////////////////////
// summary :
// return  :
//////////////////////////////////////////////////////////////////////
JBErrorCode CJBLib::RecvFlashFileData(
	FlashFileData* fileData
)
{
	unsigned long data[5];
	ssize_t size = ::recv(m_socket, (char*)data, sizeof(data), 0);
	if(size !=	sizeof(data)){
		return JTAG_RECV_SOCKET_FAILED;
	}

	fileData->littleBear = ntohl(data[0])? true : false;
	fileData->compressed = ntohl(data[1])? true : false;
	fileData->recordTime = (time_t)ntohl(data[2]);
	fileData->fileSize   = ntohl(data[3]);

	unsigned long nameLen = ntohl(data[4]);
	if(nameLen > MAX_FILE_NAME_LEN){
		return JTAG_FILE_NAME_LEN_FAILED;
	}

	// file name
        size = ::recv(m_socket, fileData->fileName, nameLen, 0);
	if((size < 0) || (nameLen != (size_t)size)){
		return JTAG_RECV_SOCKET_FAILED;
	}
	fileData->fileName[nameLen] = 0;

	return JTAG_NO_ERROR;
}

//--------------------------------------------------------------------
//====================================================================
// summary :
// return  :
//====================================================================
JBErrorCode CJBLib::DeviceListRequest(
	vector<unsigned int>* deviceList
)
{
	if(!PrepareSocket(0, NORMAL_PORT_NO)){
		return JTAG_OPEN_SOCKET_FAILED;
	}

	// send header
	if(!SendHeader(DEVICE_LIST_REQ)){
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// receive response common header
	JBErrorCode status = ReceiveResponse(DEVICE_LIST_RES);
	if(status != JTAG_NO_ERROR){
		CloseSocket();
		return status;
	}

	// receive device count
	unsigned long deviceNum;
	if(!ReceiveLongData(&deviceNum)){
		LOG("failed to receive device count\n");
		CloseSocket();
		return JTAG_RECV_SOCKET_FAILED;
	}

	// receive each device ID and make list
	unsigned int i;
	for(i = 0 ; i < deviceNum ; i++){
		unsigned long deviceID;
		if(!ReceiveLongData(&deviceID)){
			CloseSocket();
			LOG("failed to receive device ID, count : %d\n", deviceNum);
			return JTAG_RECV_SOCKET_FAILED;
		}
		deviceList->push_back(deviceID);
	}

	return JTAG_NO_ERROR;
}

//--------------------------------------------------------------------
//====================================================================
// summary : get maximum size of config data size in flash memory
// return  :
//====================================================================
JBErrorCode CJBLib::MaxConfigDataSizeRequest(
	unsigned long* maxSize
)
{
	if(!PrepareSocket(0, NORMAL_PORT_NO)){
		return JTAG_OPEN_SOCKET_FAILED;
	}

	JBErrorCode status = GetMaxDataSize(CONFIG_DATA_SIZE, maxSize);
	if(status != JTAG_NO_ERROR){
		CloseSocket();
	}
	return status;
}

//====================================================================
// summary : get maximum size of program size in flash memory
// return  :
//====================================================================
JBErrorCode CJBLib::MaxUpdateProgramSizeRequest(
	unsigned long* maxSize
)
{
	if(!PrepareSocket(0, NORMAL_PORT_NO)){
		return JTAG_OPEN_SOCKET_FAILED;
	}

	JBErrorCode status = GetMaxDataSize(UPDATE_PROGRAM_SIZE, maxSize);
	if(status != JTAG_NO_ERROR){
		CloseSocket();
	}
	return status;
}

//////////////////////////////////////////////////////////////////////
// summary :
// return  :
//////////////////////////////////////////////////////////////////////
JBErrorCode CJBLib::GetMaxDataSize(
	MaxSizeType type,
	unsigned long* maxSize
)
{
	// send header
	if(!SendHeader(MEMORY_SIZE_REQ)){
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// send type
	if(!SendLongData((unsigned long)type)){
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// receive response common header
	JBErrorCode status = ReceiveResponse(MEMORY_SIZE_RES);
	if(status != JTAG_NO_ERROR){
		CloseSocket();
		return status;
	}

	// receive max size
	if(!ReceiveLongData(maxSize)){
		CloseSocket();
		return JTAG_RECV_SOCKET_FAILED;
	}

	return JTAG_NO_ERROR;
}

//--------------------------------------------------------------------
//====================================================================
// summary :
// return  :
//====================================================================
JBErrorCode CJBLib::FlashFileDataRequest(
	FlashFileData* fileData
)
{
	if(!PrepareSocket(0, NORMAL_PORT_NO)){
		return JTAG_OPEN_SOCKET_FAILED;
	}

	// send header
	if(!SendHeader(FLASH_DATA_REQ)){
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// receive response common header
	JBErrorCode status = ReceiveResponse(FLASH_DATA_RES);
	if(status != JTAG_NO_ERROR){
		CloseSocket();
		return status;
	}

	// receive file data
	status = RecvFlashFileData(fileData);
	if(status != JTAG_NO_ERROR){
		CloseSocket();
	}

	return status;
}

//--------------------------------------------------------------------
//====================================================================
// summary :
// return  :
//====================================================================
JBErrorCode CJBLib::TransferRateRequest(
	unsigned long* transferRate
)
{
	if(!PrepareSocket(0, NORMAL_PORT_NO)){
		return JTAG_OPEN_SOCKET_FAILED;
	}

	// send id
	if(!SendHeader(GET_TRANS_RATE_REQ)){
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// receive response common header
	JBErrorCode status = ReceiveResponse(GET_TRANS_RATE_RES);
	if(status != JTAG_NO_ERROR){
		CloseSocket();
		return status;
	}

	// receive current transfer rate
	if(!ReceiveLongData(transferRate)){
		CloseSocket();
		return JTAG_RECV_SOCKET_FAILED;
	}

	LOG("Transfer Rate Request, rate : %d\n", *transferRate);
	return JTAG_NO_ERROR;
}

//--------------------------------------------------------------------
//====================================================================
// summary :
// return  :
//====================================================================
JBErrorCode CJBLib::ChangeTransferRateRequest(
	unsigned long transferRate
)
{
	if(!PrepareSocket(0, NORMAL_PORT_NO)){
		return JTAG_OPEN_SOCKET_FAILED;
	}

	// send id
	if(!SendHeader(SET_TRANS_RATE_REQ)){
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// send transfer rate
	if(!SendLongData(transferRate)){
		LOG("Failed to send tranfer rate, rate : %d\n", transferRate);
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// receive response common header
	JBErrorCode status = ReceiveResponse(SET_TRANS_RATE_RES);
	if(status != JTAG_NO_ERROR){
		LOG("Failed to set rate, rate : %d\n", transferRate);
		CloseSocket();
		return status;
	}

	return JTAG_NO_ERROR;
}

//--------------------------------------------------------------------
//====================================================================
// summary :
// return  :
//====================================================================
JBErrorCode CJBLib::FirmwareVersionRequest(
	char* versionName
)
{
	if(!PrepareSocket(0, NORMAL_PORT_NO)){
		return JTAG_OPEN_SOCKET_FAILED;
	}

	// send id
	if(!SendHeader(GET_FIRM_VERSION_REQ)){
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// receive response common header
	JBErrorCode status = ReceiveResponse(GET_FIRM_VERSION_RES);
	if(status != JTAG_NO_ERROR){
		CloseSocket();
		return status;
	}

	// receive current transfer rate
	unsigned long len;
	if(!ReceiveLongData(&len)){
		CloseSocket();
		return JTAG_RECV_SOCKET_FAILED;
	}
	if(len > MAX_VERSION_NAME_LEN){
		return JTAG_VERSION_NAME_FAILED;
	}

	ssize_t size = ::recv(m_socket, versionName, len, 0);
	if((size < 0) || (len != (unsigned long)size)){
		return JTAG_RECV_SOCKET_FAILED;
	}
	versionName[len] = 0;

	LOG("Firmware version name : %s\n", versionName);
	return JTAG_NO_ERROR;
}

//////////////////////////////////////////////////////////////////////
// summary :
// return  :
//////////////////////////////////////////////////////////////////////
unsigned int CJBLib::CalculateCheckSum(
	unsigned char* data,
	size_t size
) const
{
	unsigned int sum = 0;
	unsigned int i;
	for(i = 0 ; i < size ; i++){
		sum += data[i];
	}
	return sum;
}

//////////////////////////////////////////////////////////////////////
// summary :
// return  :
//////////////////////////////////////////////////////////////////////
JBErrorCode CJBLib::SendFileData(
	const char* fileName,
	JBDataID waitID,
	JBDataID notifyID,
	void (*notifyFunc)(int status, int param)
)
{
	ASSERT(m_socket >= 0);
	ASSERT(fileName);

	JBErrorCode status = JTAG_NO_ERROR;

	// get file size
	struct stat fileStatus;
	int ret = stat(fileName, &fileStatus);
	if(ret < 0){
		LOG("get file info failed, name : %s\n", fileName);
		return JTAG_OPEN_FILE_FAILED;
	}
	unsigned long fileSize = fileStatus.st_size;

	// send fileSize
	if(!SendLongData(fileSize)){
		CloseSocket();
		return JTAG_SEND_SOCKET_FAILED;
	}

	// open indicated file to read data
	FILE* fp = fopen(fileName, "rb");
	if(fp == NULL){
		CloseSocket();
		return JTAG_OPEN_FILE_FAILED;
	}

	fd_set readfds;
	fd_set writefds;
	unsigned long cksum = 0;
	unsigned int wrote = 0;
	char buf[WRITE_BUF_SIZE];
	while(1){
		FD_ZERO(&readfds);
		FD_ZERO(&writefds);
		FD_SET(m_socket, &readfds);
		if(wrote < (fileSize + CHECK_SUM_SIZE)){
			FD_SET(m_socket, &writefds);
		}

		int ret = select(m_socket + 1, &readfds, &writefds, NULL, NULL);
		if(ret < 0){
			status = JTAG_SELECT_SOCKET_FAILED;
			goto done;
		}

		if(FD_ISSET(m_socket, &writefds)){
			////////////////////////////////////////////////////////////////
			// write enabled
			////////////////////////////////////////////////////////////////
			int opt = 0;
#ifndef WIN32
			opt = MSG_DONTWAIT;
#endif
			if(wrote == fileSize){
				// send check sum
				if(!SendLongData(cksum)){
					status = JTAG_SEND_SOCKET_FAILED;
					goto done;
				}
				wrote += CHECK_SUM_SIZE;
			}
			else{
				// send file data
				size_t redSize = fread(buf, 1, sizeof(buf), fp);
				if(redSize <= 0){
					status = JTAG_READ_FILE_FAILED;
					goto done;
				}
				int sendSize = ::send(m_socket, buf, redSize, opt);
				if(sendSize <= 0){
					status = JTAG_SEND_SOCKET_FAILED;
					goto done;
				}
				wrote += sendSize;
				cksum += CalculateCheckSum((unsigned char*)buf, redSize);
			}
		}
		if(FD_ISSET(m_socket, &readfds)){
			////////////////////////////////////////////////////////////////
			// receive enabled
			////////////////////////////////////////////////////////////////
			JBDataHeader header;
			if(!ReceiveHeader(&header)){
				status = JTAG_RECV_SOCKET_FAILED;
				goto done;
			}

			if(header.version != PROTOCOL_VERSION){
				status = JTAG_VERSION_MISMATCH;
				goto done;
			}
			if(header.sequence != m_sequence){
				status = JTAG_INVALID_RESPONSE;
				goto done;
			}

			if(header.dataID == waitID){
				if(!ReceiveLongData((unsigned long*)&status)){
					status = JTAG_RECV_SOCKET_FAILED;
					goto done;
				}
				goto done;
			}
			else if(header.dataID == notifyID){
				// inform progress data
				unsigned long notifyType;
				unsigned long param;
				if(!ReceiveLongData((unsigned long*)&notifyType)){
					status = JTAG_RECV_SOCKET_FAILED;
					goto done;
				}
				if(!ReceiveLongData(&param)){
					status = JTAG_RECV_SOCKET_FAILED;
					goto done;
				}
				if(notifyFunc){
					notifyFunc(notifyType, param);
				}
			}
			else{
				LOG("No such data ID exist : %d\n", header.dataID);
				status = JTAG_INVALID_RESPONSE;
				goto done;
			}
		}
	}

done:
	fclose(fp);
	return status;
}
